#!/usr/bin/env python
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.


import textwrap
from argparse import Namespace

from croud.gql import Query, print_query


def org_users_add(args: Namespace):
    mutation = textwrap.dedent(
        """
        mutation addUserToOrganization($input: AddUserToOrganizationInput!) {
          addUserToOrganization(input: $input) {
            user {
              uid
              email
              organizationId
            }
          }
        }
    """
    ).strip()

    vars = {"input": {"user": args.user, "organizationId": args.org_id}}
    if args.role is not None:
        vars["input"]["roleFqn"] = args.role

    query = Query(mutation, args)
    query.execute(vars)
    print_query(query, "addUserToOrganization")


def org_users_remove(args: Namespace):
    mutation = textwrap.dedent(
        """
        mutation removeUserFromOrganization($input: RemoveUserFromOrganizationInput!) {
          removeUserFromOrganization(input: $input) {
            success
          }
        }
    """
    ).strip()

    vars = {"input": {"user": args.user, "organizationId": args.org_id}}

    query = Query(mutation, args)
    query.execute(vars)
    print_query(query, "removeUserFromOrganization")



from crate.theme.rtd.conf.cloud_cli import *

exclude_patterns = ["out/**"]

# crate.theme sets html_favicon to favicon.png which causes a warning because
# it should be a .ico and in addition there is no favicon.png in this project
# so it can't find the file
html_favicon = None

site_url = 'https://crate.io/docs/croud/en/latest/'
extensions = ['sphinx_sitemap', 'sphinxarg.ext']

# The file extension
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'


html_static_path = ['_static']
html_context = {
    'extra_css_files': [
        '_static/sphinxarg.css',  # overrides
    ],
}



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from unittest import mock

from aiohttp import TCPConnector  # type: ignore
from aiohttp.test_utils import setup_test_loop, teardown_test_loop
from util.fake_server import FakeCrateDBCloud, FakeResolver

from croud.config import Configuration
from croud.rest import Client
from croud.session import HttpSession, RequestMethod


@mock.patch.object(Configuration, "get_token", return_value="eyJraWQiOiIx")
@mock.patch("croud.session.cloud_url", return_value="https://cratedb.local")
class TestRestClient:
    @classmethod
    def setup_class(cls):
        cls.loop = setup_test_loop()
        cls.fake_cloud = FakeCrateDBCloud(loop=cls.loop)
        cls.info = cls.loop.run_until_complete(cls.fake_cloud.start())
        cls.resolver = FakeResolver(cls.info, loop=cls.loop)

    @classmethod
    def teardown_class(cls):
        cls.loop.run_until_complete(cls.fake_cloud.stop())
        teardown_test_loop(cls.loop)

    def test_send_success_sets_data_with_key(self, mock_cloud_url, mock_token):
        with mock.patch.object(HttpSession, "_get_conn", return_value=self._get_conn()):
            data = {"key": "value"}

            client = Client(env="dev", region="bregenz.a1", output_fmt="json")
            client.send(RequestMethod.GET, "/data/data-key")

            assert client._data == data
            assert client._error is None

    def test_send_success_sets_data_without_key(self, mock_cloud_url, mock_token):
        with mock.patch.object(HttpSession, "_get_conn", return_value=self._get_conn()):
            client = Client(env="dev", region="bregenz.a1", output_fmt="json")
            client.send(RequestMethod.GET, "/data/no-key")

            assert client._data == {"key": "value"}
            assert client._error is None

    def test_send_error_sets_error(self, mock_cloud_url, mock_token):
        with mock.patch.object(HttpSession, "_get_conn", return_value=self._get_conn()):
            client = Client(env="dev", region="bregenz.a1", output_fmt="json")
            client.send(RequestMethod.GET, "/errors/400")

            assert client._error == {
                "message": "Bad request.",
                "errors": {"key": "Error on 'key'"},
            }

    def test_send_text_response(self, mock_cloud_url, mock_token):
        with mock.patch.object(HttpSession, "_get_conn", return_value=self._get_conn()):
            client = Client(env="dev", region="bregenz.a1", output_fmt="json")
            client.send(RequestMethod.GET, "/text-response")

            assert client._error == {
                "message": "Invalid response type.",
                "success": False,
            }

    def _get_conn(self):
        return TCPConnector(loop=self.loop, resolver=self.resolver, ssl=True)


@mock.patch("croud.rest.print_success")
@mock.patch("croud.rest.print_format")
def test_print_success(mock_print_format, mock_print_success, event_loop):
    client = Client(env="dev", region="bregenz.a1", output_fmt="json", loop=event_loop)

    client._data = {"key": "value"}
    client.print()
    mock_print_format.assert_called_once_with({"key": "value"}, "json", None)

    client._data = None
    client.print("Success message.")
    client.print()

    mock_print_success.assert_has_calls(
        [mock.call("Success message."), mock.call("Success.")]
    )


@mock.patch("croud.rest.print_error")
def test_print_error(mock_print_error, event_loop):
    client = Client(env="dev", region="bregenz.a1", output_fmt="json", loop=event_loop)

    error = {"message": "Bad request.", "errors": {"key": "Error on 'key'"}}
    client._error = error
    client.print()
    mock_print_error.assert_called_once_with("Bad request.")


@mock.patch("croud.rest.print_format")
def test_print_error_no_message(mock_print_format, event_loop):
    client = Client(env="dev", region="bregenz.a1", output_fmt="json", loop=event_loop)

    error = {"errors": {"key": "Error on 'key'"}}
    client._error = error
    client.print()
    mock_print_format.assert_called_once_with(error, "json")



# PYTHON_ARGCOMPLETE_OK
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import sys

import colorama

from croud.clusters.commands import clusters_deploy, clusters_list
from croud.cmd import (
    cluster_id_arg,
    cluster_name_arg,
    consumer_id_arg,
    consumer_name_arg,
    consumer_schema_arg,
    consumer_table_arg,
    crate_password_arg,
    crate_username_arg,
    crate_version_arg,
    eventhub_consumer_group_arg,
    eventhub_dsn_arg,
    lease_storage_container_arg,
    lease_storage_dsn_arg,
    num_instances_arg,
    org_id_arg,
    org_id_no_org_arg_mutual_exclusive,
    org_name_arg,
    org_plan_type_arg,
    product_name_arg,
    product_tier_arg,
    product_unit_arg,
    project_id_arg,
    project_name_arg,
    resource_id_arg,
    role_fqn_arg,
    user_id_arg,
    user_id_or_email_arg,
)
from croud.config import Configuration, config_get, config_set
from croud.consumers.commands import consumers_deploy, consumers_edit, consumers_list
from croud.login import login
from croud.logout import logout
from croud.me import me
from croud.monitoring.grafana.commands import set_grafana
from croud.organizations.commands import organizations_create, organizations_list
from croud.organizations.users.commands import org_users_add, org_users_remove
from croud.parser import create_parser
from croud.projects.commands import project_create, projects_list
from croud.projects.users.commands import project_user_add, project_user_remove
from croud.users.commands import users_list
from croud.users.roles.commands import roles_add, roles_list, roles_remove

# fmt: off
command_tree = {
    "me": {
        "help": "Print information about the current logged in user.",
        "extra_args": [],
        "resolver": me,
    },
    "login": {"help": "Log in to CrateDB Cloud.", "resolver": login},
    "logout": {"help": "Log out of CrateDB Cloud.", "resolver": logout},
    "config": {
        "help": "Manage croud default configuration values.",
        "commands": {
            "get": {
                "help": "Get default configuration values.",
                "resolver": config_get,
                "noop_arg": {"choices": ["env", "region", "output-fmt"]},
            },
            "set": {
                "help": "Set default configuration values.",
                "extra_args": [],
                "resolver": config_set,
            },
        },
    },
    "consumers": {
        "help": "Manage consumers.",
        "commands": {
            "deploy": {
                "help": "Deploy a new consumer.",
                "extra_args": [
                    lambda req_opt_group, opt_opt_group: product_name_arg(
                        req_opt_group, opt_opt_group, True
                    ),
                    product_tier_arg,
                    consumer_name_arg,
                    num_instances_arg,
                    lambda req_opt_group, opt_opt_group: consumer_schema_arg(
                        req_opt_group, opt_opt_group, True
                    ),
                    lambda req_opt_group, opt_opt_group: consumer_table_arg(
                        req_opt_group, opt_opt_group, True
                    ),
                    lambda req_opt_group, opt_opt_group: project_id_arg(
                        req_opt_group, opt_opt_group, True
                    ),
                    lambda req_opt_group, opt_opt_group: cluster_id_arg(
                        req_opt_group, opt_opt_group, True
                    ),
                    lambda req_opt_group, opt_opt_group: eventhub_dsn_arg(
                        req_opt_group, opt_opt_group, True
                    ),
                    lambda req_opt_group, opt_opt_group: eventhub_consumer_group_arg(
                        req_opt_group, opt_opt_group, True
                    ),
                    lambda req_opt_group, opt_opt_group: lease_storage_dsn_arg(
                        req_opt_group, opt_opt_group, True
                    ),
                    lambda req_opt_group, opt_opt_group: lease_storage_container_arg(
                        req_opt_group, opt_opt_group, True
                    ),
                ],
                "resolver": consumers_deploy,
            },
            "list": {
                "help": "List all consumers the current user has access to.",
                "extra_args": [
                    lambda req_opt_group, opt_opt_group: project_id_arg(
                        req_opt_group, opt_opt_group, False
                    ),
                    lambda req_opt_group, opt_opt_group: cluster_id_arg(
                        req_opt_group, opt_opt_group, False
                    ),
                    lambda req_opt_group, opt_opt_group: product_name_arg(
                        req_opt_group, opt_opt_group, False
                    )
                ],
                "resolver": consumers_list,
            },
            "edit": {
                "help": "Edit the specified consumer set.",
                "extra_args": [
                    consumer_id_arg,
                    lambda req_opt_group, opt_opt_group:
                    eventhub_dsn_arg(
                        req_opt_group, opt_opt_group, False
                    ),
                    lambda req_opt_group, opt_opt_group:
                    eventhub_consumer_group_arg(
                        req_opt_group, opt_opt_group, False
                    ),
                    lambda req_opt_group, opt_opt_group:
                    lease_storage_dsn_arg(
                        req_opt_group, opt_opt_group, False
                    ),
                    lambda req_opt_group, opt_opt_group:
                    lease_storage_container_arg(
                        req_opt_group, opt_opt_group, False
                    ),
                    lambda req_opt_group, opt_opt_group:
                    consumer_schema_arg(
                        req_opt_group, opt_opt_group, False
                    ),
                    lambda req_opt_group, opt_opt_group:
                    consumer_table_arg(
                        req_opt_group, opt_opt_group, False
                    ),
                    lambda req_opt_group, opt_opt_group:
                    cluster_id_arg(
                        req_opt_group, opt_opt_group, False
                    ),
                ],
                "resolver": consumers_edit,
            },
        },
    },
    "projects": {
        "help": "Manage projects.",
        "commands": {
            "create": {
                "help": "Create a project in the specified organization and region.",
                "extra_args": [
                    project_name_arg,
                    lambda req_opt_group, opt_opt_group: org_id_arg(
                        req_opt_group, opt_opt_group, True
                    ),
                ],
                "resolver": project_create,
            },
            "list": {
                "help": (
                    "List all projects the current user has access to in "
                    "the specified region."
                ),
                "extra_args": [],
                "resolver": projects_list,
            },
            "users": {
                "help": "Manage users in projects.",
                "commands": {
                    "add": {
                        "help": "Add the selected user to a project.",
                        "extra_args": [
                            lambda req_opt_group, opt_opt_group: project_id_arg(
                                req_opt_group, opt_opt_group, True
                            ),
                            user_id_or_email_arg,
                        ],
                        "resolver": project_user_add,
                    },
                    "remove": {
                        "help": "Remove the selected user from a project.",
                        "extra_args": [
                            lambda req_opt_group, opt_opt_group: project_id_arg(
                                req_opt_group, opt_opt_group, True
                            ),
                            user_id_or_email_arg,
                        ],
                        "resolver": project_user_remove,
                    },
                },
            },
        },
    },
    "clusters": {
        "help": "Manage clusters.",
        "commands": {
            "list": {
                "help": "List all clusters the current user has access to.",
                "extra_args": [
                    lambda req_opt_group, opt_opt_group: project_id_arg(
                        req_opt_group, opt_opt_group, False
                    ),
                ],
                "resolver": clusters_list,
            },
            "deploy": {
                "help": "Deploy a new CrateDB cluster.",
                "extra_args": [
                    lambda req_opt_group, opt_opt_group: product_name_arg(
                        req_opt_group, opt_opt_group, True
                    ),
                    product_tier_arg,
                    lambda req_opt_group, opt_opt_group: product_unit_arg(
                        req_opt_group, opt_opt_group, False
                    ),
                    lambda req_opt_group, opt_opt_group: project_id_arg(
                        req_opt_group, opt_opt_group, True
                    ),
                    lambda req_opt_group, opt_opt_group: cluster_name_arg(
                        req_opt_group, opt_opt_group, True
                    ),
                    crate_version_arg,
                    crate_username_arg,
                    crate_password_arg,
                ],
                "resolver": clusters_deploy,
            }
        },
    },
    "organizations": {
        "help": "Manage organizations.",
        "commands": {
            "create": {
                "help": "Create a new organization.",
                "extra_args": [org_name_arg, org_plan_type_arg],
                "resolver": organizations_create,
            },
            "list": {
                "help": "List all organizations the current user has access to.",
                "extra_args": [],
                "resolver": organizations_list,
            },
            "users": {
                "help": "Manage users in an organization.",
                "commands": {
                    "add": {
                        "help": "Add the selected user to an organization.",
                        "extra_args": [
                            user_id_or_email_arg,
                            lambda req_opt_group, opt_opt_group: role_fqn_arg(
                                req_opt_group, opt_opt_group, False
                            ),
                            lambda req_opt_group, opt_opt_group: org_id_arg(
                                req_opt_group, opt_opt_group, False
                            ),
                        ],
                        "resolver": org_users_add,
                    },
                    "remove": {
                        "help": "Remove the selected user from an organization.",
                        "extra_args": [
                            user_id_or_email_arg,
                            lambda req_opt_group, opt_opt_group: org_id_arg(
                                req_opt_group, opt_opt_group, False
                            ),
                        ],
                        "resolver": org_users_remove,
                    },
                },
            },
        },
    },
    "users": {
        "help": "Manage users.",
        "commands": {
            "list": {
                "help": "List all users within the specified organization.",
                "extra_args": [
                    org_id_no_org_arg_mutual_exclusive,
                ],
                "resolver": users_list,
            },
            "roles": {
                "help": "Manage user roles.",
                "commands": {
                    "add": {
                        "help": "Assign a role to the selected user.",
                        "extra_args": [
                            lambda req_opt_group, opt_opt_group: resource_id_arg(
                                req_opt_group, opt_opt_group, True
                            ),
                            lambda req_opt_group, opt_opt_group: user_id_arg(
                                req_opt_group, opt_opt_group, True
                            ),
                            lambda req_opt_group, opt_opt_group: role_fqn_arg(
                                req_opt_group, opt_opt_group, True
                            ),
                        ],
                        "resolver": roles_add,
                    },
                    "remove": {
                        "help": "Unassign a role from the selected user.",
                        "extra_args": [
                            lambda req_opt_group, opt_opt_group: resource_id_arg(
                                req_opt_group, opt_opt_group, True
                            ),
                            lambda req_opt_group, opt_opt_group: user_id_arg(
                                req_opt_group, opt_opt_group, True
                            ),
                            lambda req_opt_group, opt_opt_group: role_fqn_arg(
                                req_opt_group, opt_opt_group, True
                            ),
                        ],
                        "resolver": roles_remove,
                    },
                    "list": {
                        "help": "List all available roles.",
                        "extra_args": [],
                        "resolver": roles_list,
                    },
                },
            },
        },
    },
    "monitoring": {
        "help": "Manage monitoring tools.",
        "commands": {
            "grafana": {
                "help": "Manage access to Grafana dashboards for projects.",
                "commands": {
                    "enable": {
                        "help": "Enable Grafana dashboards to visualize metrics for a "
                        "project.",
                        "extra_args": [
                            lambda req_opt_group, opt_opt_group: project_id_arg(
                                req_opt_group, opt_opt_group, True
                            ),
                        ],
                        "resolver": lambda args: set_grafana(True, args),
                    },
                    "disable": {
                        "help": "Disable Grafana dashboards for a project.",
                        "extra_args": [
                            lambda req_opt_group, opt_opt_group: project_id_arg(
                                req_opt_group, opt_opt_group, True
                            ),
                        ],
                        "resolver": lambda args: set_grafana(False, args),
                    },
                },
            },
        }
    }
}
# fmt: on


def get_parser():
    tree = {
        "help": "A command line interface for CrateDB Cloud.",
        "commands": command_tree,
    }
    return create_parser(tree)


def main():
    Configuration.create()
    colorama.init()

    parser = get_parser()
    params = parser.parse_args(sys.argv[1:])
    if "resolver" in params:
        fn = params.resolver
        del params.resolver
        fn(params)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import asyncio
from argparse import Namespace
from typing import Optional

from croud.config import Configuration
from croud.printer import print_error, print_info
from croud.server import Server
from croud.session import cloud_url
from croud.util import can_launch_browser, open_page_in_browser

LOGIN_PATH = "/oauth2/login?cli=true"


def login(args: Namespace) -> None:
    """
    Performs an OAuth2 Login to CrateDB Cloud
    """

    if can_launch_browser():
        loop = asyncio.get_event_loop()
        server = Server(loop)
        server.create_web_app()
        loop.run_until_complete(server.start())

        env = _set_login_env(args.env)
        open_page_in_browser(_login_url(env))
        print_info("A browser tab has been launched for you to login.")

        try:
            loop.run_forever()
        except KeyboardInterrupt:
            loop.run_until_complete(server.stop())
            exit(1)
        finally:
            loop.run_until_complete(server.stop())
        loop.close()

        Configuration.set_context(env.lower())
    else:
        print_error("Login only works with a valid browser installed.")
        exit(1)

    print_info("Login successful.")


def _set_login_env(env: Optional[str]) -> str:
    env = Configuration.get_env() if env is None else env
    Configuration.override_context(env)
    return env


def _login_url(env: str) -> str:
    return cloud_url(env.lower()) + LOGIN_PATH



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import pytest

from croud.printer import FormatPrinter


class TestFormatPrinter:
    @pytest.mark.parametrize(
        "rows,expected",
        (
            (None, "null\n"),
            ({}, "{}\n"),
            (
                {"a": "foo", "b": 1, "c": True, "d": {"x": 1, "y": "xyz", "z": False}},
                (
                    "{\n"
                    '  "a": "foo",\n'
                    '  "b": 1,\n'
                    '  "c": true,\n'
                    '  "d": {\n'
                    '    "x": 1,\n'
                    '    "y": "xyz",\n'
                    '    "z": false\n'
                    "  }\n"
                    "}\n"
                ),
            ),
        ),
    )
    def test_json_format(self, rows, expected, capsys):
        FormatPrinter().print_rows(rows, "json")
        out, err = capsys.readouterr()
        assert out == expected

    @pytest.mark.parametrize(
        "rows,keys,expected",
        (
            (None, ["what", "ever"], "\n"),
            ({}, ["what", "ever"], "\n"),
            (
                {"a": "foo", "b": 1, "c": True},
                None,
                (
                    "+-----+-----+------+\n"
                    "| a   |   b | c    |\n"
                    "|-----+-----+------|\n"
                    "| foo |   1 | TRUE |\n"
                    "+-----+-----+------+\n"
                ),
            ),
            (
                [
                    {"a": "foo", "b": 1, "c": True},
                    {"b": 2, "c": False, "a": {"bar": [{"value": 0}, {"value": 1}]}},
                ],
                None,
                (
                    "+---------------------------------------+-----+-------+\n"
                    "| a                                     |   b | c     |\n"
                    "|---------------------------------------+-----+-------|\n"
                    "| foo                                   |   1 | TRUE  |\n"
                    '| {"bar": [{"value": 0}, {"value": 1}]} |   2 | FALSE |\n'
                    "+---------------------------------------+-----+-------+\n"
                ),
            ),
            (
                {"k1": "v1", "k2": "v2", "k3": "v3"},
                ["k1", "k3"],
                (
                    "+------+------+\n"
                    "| k1   | k3   |\n"
                    "|------+------|\n"
                    "| v1   | v3   |\n"
                    "+------+------+\n"
                ),
            ),
            (
                [
                    {"k1": "v01", "k2": "v02", "k3": "v03"},
                    {"k1": "v11", "k2": "v12", "k3": "v13"},
                ],
                ["k1", "k2"],
                (
                    "+------+------+\n"
                    "| k1   | k2   |\n"
                    "|------+------|\n"
                    "| v01  | v02  |\n"
                    "| v11  | v12  |\n"
                    "+------+------+\n"
                ),
            ),
            (
                [
                    {"k1": "v01", "k2": "v02", "k3": "v03"},
                    {"k1": "v11", "k2": "v12", "k3": "v13"},
                ],
                ["k1", "k2", "k4"],
                (
                    "+------+------+\n"
                    "| k1   | k2   |\n"
                    "|------+------|\n"
                    "| v01  | v02  |\n"
                    "| v11  | v12  |\n"
                    "+------+------+\n"
                ),
            ),
        ),
    )
    def test_tabular_format(self, rows, keys, expected, capsys):
        FormatPrinter(keys=keys).print_rows(rows, "table")
        out, err = capsys.readouterr()
        assert out == expected



#!/usr/bin/env python
#
# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import asyncio
from argparse import Namespace
from unittest.mock import patch

import pytest

from croud.config import Configuration
from croud.gql import Query, print_query


@pytest.mark.parametrize(
    "input,expected", [(Namespace(env="dev"), "dev"), (Namespace(env=None), "prod")]
)
@patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
def test_env_set(mock_load_config, input, expected):
    query = Query("{}", input)
    assert query._env == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (Namespace(env="dev", output_fmt="json"), "json"),
        (Namespace(env="dev", output_fmt=None), "table"),
        (Namespace(env="dev"), "table"),
    ],
)
@patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
def test_output_fmt_set(mock_load_config, input, expected):
    query = Query("{}", input)
    assert query._output_fmt == expected


@pytest.mark.parametrize(
    "input,expected",
    [
        (Namespace(env="dev", region="westeurope.azure"), "westeurope.azure"),
        (Namespace(env="dev", region=None), "bregenz.a1"),
        (Namespace(env="dev"), "bregenz.a1"),
    ],
)
@patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
def test_region_set(mock_load_config, input, expected):
    query = Query("{}", input)
    assert query._region == expected


@pytest.mark.parametrize(
    "response, key, message, expected_message",
    [
        (
            {"addUserToProject": {"success": True}},
            "addUserToProject",
            "User added to Project.",
            "User added to Project.",
        ),
        ({"addUserToProject": {"success": True}}, "addUserToProject", None, "Success."),
    ],
)
@patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
@patch("croud.gql.print_success")
def test_print_query_success_with_message(
    print_success, load_config, response, key, message, expected_message
):
    query = Query("", Namespace(env="test"))
    query._response = response

    print_query(query, key, message)
    print_success.assert_called_once_with(expected_message)


@pytest.mark.parametrize(
    "response, key, expected_message",
    [
        (
            {"addUserToProject": {"success": False}},
            "addUserToProject",
            "Command not successful, however no server-side errors occurred. "
            "Please try again.",
        )
    ],
)
@patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
@patch("croud.gql.print_error")
def test_print_query_not_successful(
    print_error, load_config, response, key, expected_message
):
    query = Query("", Namespace(env="test"))
    query._response = response

    print_query(query, key)
    print_error.assert_called_once_with(expected_message)


@patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
@patch("croud.gql.print_error")
def test_print_query_error(print_error, load_config):
    query = Query("", Namespace(env="test"))
    query._error = "This is a GraphQL error message"

    print_query(query)
    print_error.assert_called_once_with(query._error)


@patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
@patch("croud.gql.print_info")
def test_print_query_no_response(print_info, load_config):
    query = Query("", Namespace(env="test"))
    query._response = {}

    print_query(query)
    expected_message = "Result contained no data to print."
    print_info.assert_called_once_with(expected_message)


@patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
@patch("croud.gql.print_error")
def test_print_query_malformatted_response(print_error, load_config):
    query = Query("", Namespace(env="test"))
    query._response = 42

    print_query(query)
    expected_message = "Result has no proper format to print."
    print_error.assert_called_once_with(expected_message)


@pytest.mark.parametrize("format", ["json", "tabular"])
@pytest.mark.parametrize(
    ["response", "key"],
    [
        ({"data": ["foo", "bar"]}, None),
        ({"allNames": {"data": ["foo", "bar"]}}, "allNames"),
        ({"allNames": ["foo", "bar"]}, "allNames"),
    ],
)
@patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
@patch("croud.gql.print_format")
def test_print_query_data(print_format, load_config, response, key, format):
    query = Query("", Namespace(env="test", output_fmt=format))
    query._response = response

    print_query(query, key)
    print_format.assert_called_once_with(["foo", "bar"], format)


@pytest.mark.parametrize(
    ["response", "key"],
    [
        ({"data": []}, None),
        ({"allNames": {"data": []}}, "allNames"),
        ({"allNames": []}, "allNames"),
    ],
)
@patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
@patch("croud.gql.print_info")
def test_print_query_no_data(print_info, load_config, response, key):
    query = Query("", Namespace(env="test"))
    query._response = response

    print_query(query, key)
    expected_message = "Result contained no data to print."
    print_info.assert_called_once_with(expected_message)


@patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
def test_execute_query_with_variables(load_config):
    body = """
        query {
          me {
            uid
          }
        }
    """
    vars = {"a": "foo", "b": 42}

    result_future = asyncio.Future()
    result_future.set_result({"data": []})

    query = Query(body, Namespace(env="test"))
    with patch.object(query, "_fetch_data", return_value=result_future) as fetch_data:
        query.execute(vars)

    fetch_data.assert_called_once_with(body, vars)



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.


from unittest import mock

from aiohttp.test_utils import TestClient, TestServer, loop_context

from croud.config import Configuration
from croud.server import Server


class TestLocalServer(object):
    @mock.patch.object(Configuration, "set_token")
    def test_token_handler_and_login(self, mock_write_token):
        with loop_context() as loop:
            server = Server(loop)
            app = server.create_web_app()
            client = TestClient(TestServer(app), loop=loop)
            loop.run_until_complete(client.start_server())

            async def test_get_route():
                with mock.patch.object(loop, "stop"):
                    resp = await client.get("?token=123")
                    assert resp.status == 200
                    text = await resp.text()
                    assert "You have successfully logged into CrateDB Cloud!" in text

            loop.run_until_complete(test_get_route())
            loop.run_until_complete(client.close())

    @mock.patch.object(Configuration, "set_token")
    def test_token_missing_for_login(self, mock_write_token):
        with loop_context() as loop:
            server = Server(loop)
            app = server.create_web_app()
            client = TestClient(TestServer(app), loop=loop)
            loop.run_until_complete(client.start_server())

            async def test_get_route():
                with mock.patch.object(loop, "stop"):
                    resp = await client.get("/")
                    assert resp.status == 500
                    text = await resp.text()
                    assert "No query param 'token' in request" in text

            loop.run_until_complete(test_get_route())
            loop.run_until_complete(client.close())



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import unittest
from unittest import mock

import pytest

from croud.util import (
    can_launch_browser,
    clean_dict,
    get_platform_info,
    is_wsl,
    open_page_in_browser,
)


class TestUtils(unittest.TestCase):
    authd_mutation = {"mutationName": {"prop1": "propvalue"}}
    unauthd_mutation = {"errors": [{"message": "An error message"}]}
    authd_entity_list = {"entityListName": {"data": [{"key": "value"}]}}

    # This function was copied from the <https://github.com/Azure/azure-cli>
    # project. See `LICENSE` for more information.
    @mock.patch("webbrowser.open", autospec=True)
    @mock.patch("subprocess.call", autospec=True)
    def test_open_page_in_browser(self, subprocess_call_mock, webbrowser_open_mock):
        platform_name, release = get_platform_info()
        open_page_in_browser("http://foo")
        if is_wsl(platform_name, release):
            subprocess_call_mock.assert_called_once_with(
                ["cmd.exe", "/c", "start http://foo"]
            )
        else:
            webbrowser_open_mock.assert_called_once_with("http://foo", 2)

    # This function was copied from the <https://github.com/Azure/azure-cli>
    # project. See `LICENSE` for more information.
    @mock.patch("croud.util.get_platform_info", autospec=True)
    @mock.patch("webbrowser.get", autospec=True)
    def test_can_launch_browser(self, webbrowser_get_mock, get_platform_mock):
        # WSL is always fine
        get_platform_mock.return_value = ("linux", "4.4.0-17134-microsoft")
        result = can_launch_browser()
        self.assertTrue(result)

        # windows is always fine
        get_platform_mock.return_value = ("windows", "10")
        result = can_launch_browser()
        self.assertTrue(result)

        # osx is always fine
        get_platform_mock.return_value = ("darwin", "10")
        result = can_launch_browser()
        self.assertTrue(result)

        # now tests linux
        with mock.patch("os.environ", autospec=True) as env_mock:
            # when no GUI, return false
            get_platform_mock.return_value = ("linux", "4.15.0-1014-azure")
            env_mock.get.return_value = None
            result = can_launch_browser()
            self.assertFalse(result)

            # when there is gui, and browser is a good one, return True
            browser_mock = mock.MagicMock()
            browser_mock.name = "goodone"
            env_mock.get.return_value = "foo"
            result = can_launch_browser()
            self.assertTrue(result)

            # when there is gui, but the browser is text mode, return False
            browser_mock = mock.MagicMock()
            browser_mock.name = "www-browser"
            webbrowser_get_mock.return_value = browser_mock
            env_mock.get.return_value = "foo"
            result = can_launch_browser()
            self.assertFalse(result)


@pytest.mark.parametrize(
    ["raw", "cleaned"],
    [
        ({}, {}),
        (
            {"int": 0, "str": "", "none": None, "list": []},
            {"int": 0, "str": "", "list": []},
        ),
        (
            {"empty": {}, "nested": {"value": 42, "none": None}},
            {"empty": {}, "nested": {"value": 42}},
        ),
    ],
)
def test_clean_dict(raw, cleaned):
    assert clean_dict(raw) == cleaned



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import asyncio
import pathlib
import socket
import ssl
from typing import Any, Dict, Iterable

from aiohttp import web
from aiohttp.resolver import DefaultResolver
from aiohttp.test_utils import unused_port


class FakeResolver:
    _LOCAL_HOST = {0: "127.0.0.1", socket.AF_INET: "127.0.0.1", socket.AF_INET6: "::1"}

    def __init__(self, fakes, loop: asyncio.AbstractEventLoop):
        """fakes -- dns -> port dict"""
        self._fakes = fakes
        self._resolver = DefaultResolver(loop=loop)

    async def resolve(
        self, host: str, port: int = 0, family: socket.AddressFamily = socket.AF_INET
    ) -> Iterable[Dict[str, Any]]:
        fake_port = self._fakes.get(host)
        if fake_port is not None:
            return [
                {
                    "hostname": host,
                    "host": self._LOCAL_HOST[family],
                    "port": fake_port,
                    "family": family,
                    "proto": 0,
                    "flags": socket.AI_NUMERICHOST,
                }
            ]
        else:
            return await self._resolver.resolve(host, port, family)


class FakeCrateDBCloud:
    def __init__(self, loop: asyncio.AbstractEventLoop):
        self.loop = loop
        self.app = web.Application()
        # thi will allow us to register multiple endpoints/handlers to test
        self.app.router.add_routes([web.post("/graphql", self.on_graphql)])
        self.app.router.add_routes([web.get("/data/data-key", self.data_data_key)])
        self.app.router.add_routes([web.get("/data/no-key", self.data_no_key)])
        self.app.router.add_routes([web.get("/errors/400", self.error_400)])
        self.app.router.add_routes([web.get("/text-response", self.text_response)])

        here = pathlib.Path(__file__)
        # Load certificates and sign key used to simulate ssl/tls
        ssl_cert = here.parent / "server.crt"
        ssl_key = here.parent / "server.key"
        self.ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
        self.ssl_context.load_cert_chain(str(ssl_cert), str(ssl_key))

    async def start(self) -> Dict[str, int]:
        port = unused_port()
        self.runner = web.AppRunner(self.app)
        await self.runner.setup()
        site = web.TCPSite(self.runner, "localhost", port, ssl_context=self.ssl_context)
        await site.start()
        return {"cratedb.local": port}

    async def stop(self) -> None:
        await self.runner.cleanup()

    async def on_graphql(self, request: web.Request) -> web.Response:
        if self._is_authorized(request):
            if self._get_query_header(request) == "me":
                return web.json_response(
                    {
                        "data": {
                            "me": {
                                "email": "sheldon@crate.io",
                                "username": "Google_1234",
                                "name": "Sheldon Cooper",
                            }
                        }
                    }
                )
            resp = {"data": {"message": "Bad request"}}
            return web.json_response(resp, status=400)
        return web.Response(status=302)

    async def data_data_key(self, request: web.Request) -> web.Response:
        if self._is_authorized(request):
            return web.json_response({"data": {"key": "value"}})
        return web.Response(status=302)

    async def data_no_key(self, request: web.Request) -> web.Response:
        if self._is_authorized(request):
            return web.json_response({"key": "value"})
        return web.Response(status=302)

    async def error_400(self, request: web.Request) -> web.Response:
        if self._is_authorized(request):
            return web.json_response(
                {"message": "Bad request.", "errors": {"key": "Error on 'key'"}},
                status=400,
            )
        return web.Response(status=302)

    async def text_response(self, request: web.Request) -> web.Response:
        if self._is_authorized(request):
            return web.Response(body="Non JSON response.", status=500)
        return web.Response(status=302)

    def _is_authorized(self, request: web.Request) -> bool:
        if "session" in request.cookies:
            if request.cookies["session"]:
                return True
        return False

    def _get_query_header(self, request: web.Request) -> str:
        if request.content_type == "application/json":
            if "query" in request.headers:
                if request.headers["query"] == "me":
                    return "me"
                else:
                    return ""
        return ""



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import asyncio

from aiohttp import web

from croud.config import Configuration


class Server:
    LOGIN_MSG: str = """
    <b>You have successfully logged into CrateDB Cloud!</b>
    <p>This window can be closed.</p>"""

    PORT: int = 8400

    def __init__(self, loop: asyncio.AbstractEventLoop):
        self.loop = loop

    def create_web_app(self) -> web.Application:
        app = web.Application()
        app.add_routes([web.get("/", self.handle_session)])
        app.on_response_prepare.append(self.after_request)  # type: ignore
        self.runner = web.AppRunner(app)
        return app

    def handle_session(self, req: web.Request) -> web.Response:
        """Token handler that receives the session token from query param"""
        try:
            Configuration.set_token(req.rel_url.query["token"])
        except KeyError as ex:
            return web.Response(status=500, text=f"No query param {ex!s} in request")
        return web.Response(status=200, text=Server.LOGIN_MSG, content_type="text/html")

    async def after_request(self, req: web.Request, resp: web.Response) -> web.Response:
        """middleware callback right after a request got handled"""
        # stop the event loop right after successful login response
        # so that the Server can be terminated gracefully afterwards
        self.loop.stop()
        return resp

    async def start(self) -> None:
        """Start a local HTTP server"""
        await self.runner.setup()
        site = web.TCPSite(self.runner, "localhost", Server.PORT)
        await site.start()

    async def stop(self) -> None:
        """Shutdown the server gracefully"""
        await self.runner.cleanup()



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import textwrap
from argparse import Namespace

from croud.gql import Query, print_query
from croud.util import clean_dict


def users_list(args: Namespace) -> None:
    """
    List all users within organizations that the logged in user is part of
    """

    body = textwrap.dedent(
        """
        query allUsers($queryArgs: UserQueryArgs) {
            allUsers(sort: EMAIL, queryArgs: $queryArgs) {
                data {
                    uid
                    email
                    username
                }
            }
        }
    """
    ).strip()

    vars = clean_dict(
        {"queryArgs": {"noOrg": args.no_org, "organizationId": args.org_id}}
    )

    query = Query(body, args)
    query.execute(vars)
    print_query(query, "allUsers")



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import functools
import json
from typing import List, Optional, Union

from colorama import Fore, Style
from tabulate import tabulate

from croud.typing import JsonDict


def print_format(
    rows: Union[List[JsonDict], JsonDict], format: str, keys: Optional[List[str]] = None
) -> None:
    printer = FormatPrinter(keys)
    printer.print_rows(rows, format)


def print_error(text: str):
    print(Fore.RED + "==> Error: " + Style.RESET_ALL + text)


def print_info(text: str):
    print(Fore.CYAN + "==> Info: " + Style.RESET_ALL + text)


def print_success(text: str):
    print(Fore.GREEN + "==> Success: " + Style.RESET_ALL + text)


class FormatPrinter:
    def __init__(self, keys: Optional[List[str]] = None):
        self.supported_formats: dict = {"json": self._json, "table": self._tabular}
        self.keys = keys

    def print_rows(self, rows: Union[List[JsonDict], JsonDict], format: str) -> None:
        try:
            print_method = self.supported_formats[format]
        except KeyError as e:
            print_error(f"This print method is not supported: {e!s}")
            exit(1)
        print(print_method(rows))

    def _transform_field(self, field):
        """transform field for displaying"""
        if isinstance(field, (list, dict)):
            return json.dumps(field, sort_keys=True, ensure_ascii=False)
        elif isinstance(field, bool):
            return "TRUE" if field else "FALSE"
        else:
            return field

    def _filter_record(self, data: dict, keys: List[str]) -> JsonDict:
        return {key: value for key, value in data.items() if key in keys}

    def _json(self, rows: Union[List[JsonDict], JsonDict]) -> str:
        return json.dumps(rows, sort_keys=False, indent=2)

    def _tabular(self, rows: Union[List[JsonDict], JsonDict]) -> str:
        if rows is None:
            return ""

        if not isinstance(rows, list):
            rows = [rows]

        if self.keys:
            filter_record = functools.partial(self._filter_record, keys=self.keys)
            rows = list(map(filter_record, rows))

        # ensure that keys are mapped to their values
        # e.g. Rows, generated with
        # [{"a": "foo", "b": 1}, {"b": 2, "a": "bar"}]
        # should get printed same as
        # [{"a": "foo", "b": 1}, {"a": "bar", "b": "1"}]
        # +-----+-----+
        # | a   |   b |
        # |-----+-----|
        # | foo |   1 |
        # |-----+-----|
        # | bar |   2 |
        # +-----+-----+
        headers = list(map(str, rows[0].keys()))
        values = [
            [self._transform_field(row[header]) for header in headers] for row in rows
        ]
        return tabulate(values, headers=headers, tablefmt="psql", missingval="NULL")



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import asyncio
from argparse import Namespace

from croud.config import Configuration
from croud.printer import print_info
from croud.session import HttpSession, cloud_url

LOGOUT_PATH = "/oauth2/logout"


def logout(args: Namespace) -> None:
    if args.env is not None:
        Configuration.override_context(args.env)

    loop = asyncio.get_event_loop()
    env = Configuration.get_env()
    token = Configuration.get_token()

    loop.run_until_complete(make_request(env, token))
    Configuration.set_token("")

    print_info("You have been logged out.")


async def make_request(env: str, token: str) -> None:
    async with HttpSession(env, token) as session:
        await session.logout(_logout_url(env))


def _logout_url(env: str) -> str:
    return cloud_url(env) + LOGOUT_PATH



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import ssl
from enum import Enum
from types import TracebackType
from typing import Callable, Dict, Optional, Type

import certifi
from aiohttp import ClientResponse, ClientSession, TCPConnector  # type: ignore

from croud.printer import print_error

CLOUD_LOCAL_URL = "http://localhost:8000"
CLOUD_DEV_DOMAIN = "cratedb-dev.cloud"
CLOUD_PROD_DOMAIN = "cratedb.cloud"


class RequestMethod(Enum):
    GET = "get"
    POST = "post"
    DELETE = "delete"
    PATCH = "patch"


class HttpSession:
    def __init__(
        self,
        env: str,
        token: str,
        region: str = "bregenz.a1",
        url: Optional[str] = None,
        conn: Optional[TCPConnector] = None,
        headers: Dict[str, str] = {},
        on_new_token: Callable[[str], None] = None,
    ) -> None:
        self.env = env
        self.token = token
        self.on_new_token = on_new_token

        if not url:
            url = cloud_url(env, region)

        self.url = url
        if conn is None:
            conn = self._get_conn()

        self.client = ClientSession(
            cookies={"session": self.token}, connector=conn, headers=headers
        )

    def _get_conn(self) -> TCPConnector:
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        return TCPConnector(ssl_context=ssl_context)

    async def fetch(
        self,
        method: RequestMethod,
        endpoint: str,
        body: dict = None,
        params: dict = None,
    ) -> ClientResponse:
        url = self.url + endpoint
        resp = await getattr(self.client, method.value)(
            url, json=body, allow_redirects=False, params=params
        )
        if resp.status == 302:  # login redirect
            print_error("Unauthorized. Use `croud login` to login to CrateDB Cloud.")
            exit(1)
        return resp

    async def logout(self, url: str):
        await self.client.get(url)

    async def __aenter__(self) -> "HttpSession":
        return self

    async def close(self) -> None:
        new_token = self.client.cookie_jar.filter_cookies(self.url).get("session")
        if new_token and self.token != new_token.value and self.on_new_token:
            self.on_new_token(new_token.value)
        await self.client.close()

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.close()


def cloud_url(env: str, region: str = "bregenz.a1") -> str:
    if env == "local":
        return CLOUD_LOCAL_URL

    host = CLOUD_DEV_DOMAIN if env == "dev" else CLOUD_PROD_DOMAIN
    return f"https://{region}.{host}"



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import textwrap
import uuid
from argparse import Namespace
from unittest import mock

import pytest
from util import CommandTestCase

from croud.config import Configuration, config_get, config_set
from croud.gql import Query
from croud.rest import Client
from croud.session import RequestMethod


def gen_uuid() -> str:
    return str(uuid.uuid4())


@mock.patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
@mock.patch.object(Client, "send")
class TestMe(CommandTestCase):
    def test_me(self, mock_send, mock_config):
        argv = ["croud", "me"]
        self.assertRest(mock_send, argv, RequestMethod.GET, "/api/v2/users/me/")


class TestConfigGet:
    @mock.patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
    @mock.patch("builtins.print", autospec=True, side_effect=print)
    def test_get_env(self, mock_print, mock_load_config):
        config_get(Namespace(get="env", output_fmt=None))
        mock_print.assert_called_once_with(
            textwrap.dedent(
                """
                +-------+
                | env   |
                |-------|
                | prod  |
                +-------+
            """
            ).strip()
        )

    @mock.patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
    @mock.patch("builtins.print", autospec=True, side_effect=print)
    def test_get_top_level_setting(self, mock_print, mock_load_config):
        config_get(Namespace(get="region", output_fmt=None))
        mock_print.assert_called_once_with(
            textwrap.dedent(
                """
                +------------+
                | region     |
                |------------|
                | bregenz.a1 |
                +------------+
            """
            ).strip()
        )


class TestConfigSet:
    @mock.patch("croud.config.write_config")
    @mock.patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
    def test_set_env(self, mock_load_config, mock_write_config):
        config = Configuration.DEFAULT_CONFIG
        config["auth"]["current_context"] = "prod"

        config_set(Namespace(env="prod"))
        mock_write_config.assert_called_once_with(config)

    @mock.patch("croud.config.write_config")
    @mock.patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
    def test_set_top_level_setting(self, mock_load_config, mock_write_config):
        config = Configuration.DEFAULT_CONFIG
        config["region"] = "eastus.azure"

        config_set(Namespace(region="eastus.azure"))
        mock_write_config.assert_called_once_with(config)

        config["region"] = "bregenz.a1"


def assert_query(mock_print, expected):
    actual = mock_print.call_args[0][0]._query
    assert actual == expected


@mock.patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
@mock.patch.object(Query, "run", return_value={"data": []})
class TestClusters(CommandTestCase):
    project_id = gen_uuid()
    expected_body = textwrap.dedent(
        """
        query allClusters($filter: [ClusterFilter]) {
            allClusters(sort: [CRATE_VERSION_DESC], filter: $filter) {
                data {
                    id
                    name
                    numNodes
                    crateVersion
                    projectId
                    username
                    fqdn
                }
            }
        }
    """
    ).strip()

    @mock.patch.object(Client, "send")
    def test_list_no_project_id(self, mock_send, mock_run, mock_load_config):
        argv = ["croud", "clusters", "list"]
        self.assertRest(
            mock_send, argv, RequestMethod.GET, "/api/v2/clusters/", params={}
        )

    @mock.patch.object(Client, "send")
    def test_list_with_project_id(self, mock_send, mock_run, mock_load_config):
        argv = ["croud", "clusters", "list", "--project-id", self.project_id]
        self.assertRest(
            mock_send,
            argv,
            RequestMethod.GET,
            "/api/v2/clusters/",
            params={"project_id": self.project_id},
        )

    @mock.patch.object(Client, "send")
    def test_deploy_cluster(self, mock_send, mock_run, mock_load_config):
        argv = [
            "croud",
            "clusters",
            "deploy",
            "--product-name",
            "cratedb.az1",
            "--tier",
            "xs",
            "--unit",
            "1",
            "--project-id",
            self.project_id,
            "--cluster-name",
            "crate_cluster",
            "--version",
            "3.2.5",
            "--username",
            "foobar",
            "--password",
            "s3cr3t!",
        ]
        self.assertRest(
            mock_send,
            argv,
            RequestMethod.POST,
            "/api/v2/clusters/",
            body={
                "crate_version": "3.2.5",
                "name": "crate_cluster",
                "password": "s3cr3t!",
                "product_name": "cratedb.az1",
                "product_tier": "xs",
                "product_unit": 1,
                "project_id": self.project_id,
                "username": "foobar",
            },
        )

    @mock.patch.object(Client, "send")
    def test_deploy_cluster_no_unit(self, mock_send, mock_run, mock_load_config):
        argv = [
            "croud",
            "clusters",
            "deploy",
            "--product-name",
            "cratedb.az1",
            "--tier",
            "xs",
            "--project-id",
            self.project_id,
            "--cluster-name",
            "crate_cluster",
            "--version",
            "3.2.5",
            "--username",
            "foobar",
            "--password",
            "s3cr3t!",
        ]
        self.assertRest(
            mock_send,
            argv,
            RequestMethod.POST,
            "/api/v2/clusters/",
            body={
                "crate_version": "3.2.5",
                "name": "crate_cluster",
                "password": "s3cr3t!",
                "product_name": "cratedb.az1",
                "product_tier": "xs",
                "project_id": self.project_id,
                "username": "foobar",
            },
        )


@mock.patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
@mock.patch.object(Query, "run", return_value={"data": []})
class TestOrganizations(CommandTestCase):
    @mock.patch.object(Client, "send")
    def test_create(self, mock_send, mock_run, mock_load_config):
        argv = [
            "croud",
            "organizations",
            "create",
            "--name",
            "test-org",
            "--plan-type",
            "1",
        ]
        self.assertRest(
            mock_send,
            argv,
            RequestMethod.POST,
            "/api/v2/organizations/",
            body={"name": "test-org", "plan_type": 1},
        )

    @mock.patch.object(Client, "send")
    def test_list(self, mock_send, mock_run, mock_load_config):
        argv = ["croud", "organizations", "list"]
        self.assertRest(mock_send, argv, RequestMethod.GET, "/api/v2/organizations/")

    def test_add_user(self, mock_run, mock_load_config):
        expected_body = textwrap.dedent(
            """
            mutation addUserToOrganization($input: AddUserToOrganizationInput!) {
              addUserToOrganization(input: $input) {
                user {
                  uid
                  email
                  organizationId
                }
              }
            }
        """
        ).strip()
        expected_vars = {"input": {"user": "test@crate.io", "organizationId": None}}

        argv = ["croud", "organizations", "users", "add", "--user", "test@crate.io"]
        self.assertGql(mock_run, argv, expected_body, expected_vars)

    def test_add_user_fqn(self, mock_run, mock_load_config):
        expected_body = textwrap.dedent(
            """
            mutation addUserToOrganization($input: AddUserToOrganizationInput!) {
              addUserToOrganization(input: $input) {
                user {
                  uid
                  email
                  organizationId
                }
              }
            }
        """
        ).strip()
        expected_vars = {
            "input": {
                "roleFqn": "org_admin",
                "user": "test@crate.io",
                "organizationId": None,
            }
        }

        argv = [
            "croud",
            "organizations",
            "users",
            "add",
            "--user",
            "test@crate.io",
            "--role",
            "org_admin",
        ]
        self.assertGql(mock_run, argv, expected_body, expected_vars)

    def test_add_user_org_id(self, mock_run, mock_load_config):
        expected_body = textwrap.dedent(
            """
            mutation addUserToOrganization($input: AddUserToOrganizationInput!) {
              addUserToOrganization(input: $input) {
                user {
                  uid
                  email
                  organizationId
                }
              }
            }
        """
        ).strip()
        org_id = str(uuid.uuid4())
        expected_vars = {"input": {"organizationId": org_id, "user": "test@crate.io"}}

        argv = [
            "croud",
            "organizations",
            "users",
            "add",
            "--user",
            "test@crate.io",
            "--org-id",
            org_id,
        ]
        self.assertGql(mock_run, argv, expected_body, expected_vars)

    def test_remove_user(self, mock_run, mock_load_config):
        expected_body = textwrap.dedent(
            """
            mutation removeUserFromOrganization($input: RemoveUserFromOrganizationInput!) {
              removeUserFromOrganization(input: $input) {
                success
              }
            }
        """  # noqa
        ).strip()

        user_id = str(uuid.uuid4())
        expected_vars = {"input": {"user": user_id, "organizationId": None}}

        argv = ["croud", "organizations", "users", "remove", "--user", user_id]
        self.assertGql(mock_run, argv, expected_body, expected_vars)

    def test_remove_user_org_id(self, mock_run, mock_load_config):
        expected_body = textwrap.dedent(
            """
            mutation removeUserFromOrganization($input: RemoveUserFromOrganizationInput!) {
              removeUserFromOrganization(input: $input) {
                success
              }
            }
        """  # noqa
        ).strip()

        user_id = str(uuid.uuid4())
        org_id = str(uuid.uuid4())
        expected_vars = {"input": {"user": user_id, "organizationId": org_id}}

        argv = [
            "croud",
            "organizations",
            "users",
            "remove",
            "--user",
            user_id,
            "--org-id",
            org_id,
        ]
        self.assertGql(mock_run, argv, expected_body, expected_vars)


@mock.patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
@mock.patch.object(Client, "send")
class TestProjects(CommandTestCase):
    def test_create(self, mock_send, mock_load_config):
        argv = [
            "croud",
            "projects",
            "create",
            "--name",
            "new-project",
            "--org-id",
            "organization-id",
        ]
        self.assertRest(
            mock_send,
            argv,
            RequestMethod.POST,
            "/api/v2/projects/",
            body={"name": "new-project", "organization_id": "organization-id"},
        )

    def test_list(self, mock_send, mock_load_config):
        argv = ["croud", "projects", "list"]
        self.assertGql(mock_send, argv, RequestMethod.GET, "/api/v2/projects/")


@mock.patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
@mock.patch.object(Query, "run", return_value={"data": []})
class TestProjectsUsers(CommandTestCase):
    def test_add(self, mock_run, mock_load_config):
        expected_body = """
    mutation {
        addUserToProject(input: {
            projectId: "project-id",
            user: "user-email-or-id"
        }) {
            success
        }
    }
    """

        argv = [
            "croud",
            "projects",
            "users",
            "add",
            "--project-id",
            "project-id",
            "--user",
            "user-email-or-id",
        ]
        self.assertGql(mock_run, argv, expected_body)

    def test_remove(self, mock_run, mock_load_config):
        expected_body = """
    mutation {
        removeUserFromProject(input: {
            projectId: "project-id",
            user: "user-email-or-id"
        }) {
            success
        }
    }
    """

        argv = [
            "croud",
            "projects",
            "users",
            "remove",
            "--project-id",
            "project-id",
            "--user",
            "user-email-or-id",
        ]
        self.assertGql(mock_run, argv, expected_body)


@mock.patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
@mock.patch.object(Query, "run", return_value={"data": []})
class TestUsersRoles(CommandTestCase):
    def test_add(self, mock_run, mock_load_config):
        user_id = gen_uuid()
        role_fqn = "org_admin"
        resource_id = gen_uuid()

        expected_body = textwrap.dedent(
            """
            mutation addRoleToUser($input: UserRoleInput!) {
                addRoleToUser(input: $input) {
                    success
                }
            }
        """
        ).strip()
        expected_vars = {
            "input": {"userId": user_id, "roleFqn": role_fqn, "resourceId": resource_id}
        }

        argv = [
            "croud",
            "users",
            "roles",
            "add",
            "--user",
            user_id,
            "--role",
            role_fqn,
            "--resource",
            resource_id,
        ]
        self.assertGql(mock_run, argv, expected_body, expected_vars)

    def test_remove(self, mock_run, mock_load_config):
        user_id = gen_uuid()
        role_fqn = "org_admin"
        resource_id = gen_uuid()

        expected_body = textwrap.dedent(
            """
            mutation removeRoleFromUser($input: UserRoleInput!) {
                removeRoleFromUser(input: $input) {
                    success
                }
            }
        """
        ).strip()
        expected_vars = {
            "input": {"userId": user_id, "roleFqn": role_fqn, "resourceId": resource_id}
        }

        argv = [
            "croud",
            "users",
            "roles",
            "remove",
            "--user",
            user_id,
            "--role",
            role_fqn,
            "--resource",
            resource_id,
        ]
        self.assertGql(mock_run, argv, expected_body, expected_vars)

    def test_list(self, mock_run, mock_load_config):
        expected_body = textwrap.dedent(
            """
        query {
            allRoles {
                data {
                    fqn
                    friendlyName
                }
            }
        }
    """
        ).strip()

        argv = ["croud", "users", "roles", "list"]
        self.assertGql(mock_run, argv, expected_body)


@mock.patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
@mock.patch.object(Query, "run", return_value={"data": []})
class TestUsers(CommandTestCase):
    org_id = gen_uuid()
    expected_body = textwrap.dedent(
        """
        query allUsers($queryArgs: UserQueryArgs) {
            allUsers(sort: EMAIL, queryArgs: $queryArgs) {
                data {
                    uid
                    email
                    username
                }
            }
        }
    """
    ).strip()

    def test_list_no_filter(self, mock_run, mock_load_config):
        expected_vars = {"queryArgs": {"noOrg": False}}
        argv = ["croud", "users", "list"]
        self.assertGql(mock_run, argv, self.expected_body, expected_vars)

    def test_list_org_filter(self, mock_run, mock_load_config):
        expected_vars = {"queryArgs": {"organizationId": self.org_id, "noOrg": False}}
        argv = ["croud", "users", "list", "--org-id", self.org_id]
        self.assertGql(mock_run, argv, self.expected_body, expected_vars)

    def test_list_no_org_filter(self, mock_run, mock_load_config):
        expected_vars = {"queryArgs": {"noOrg": True}}
        argv = ["croud", "users", "list", "--no-org"]
        self.assertGql(mock_run, argv, self.expected_body, expected_vars)

    def test_list_filter_conflict(self, mock_run, mock_load_config):
        expected_vars = None
        argv = ["croud", "users", "list", "--org-id", self.org_id, "--no-org"]
        with mock.patch("sys.stderr.write") as stderr:
            with pytest.raises(SystemExit) as ex_info:
                self.assertGql(mock_run, argv, self.expected_body, expected_vars)
            stderr.assert_called_once()
            expected_error = "Argument --no-org: not allowed with argument --org-id\n\n"
            assert stderr.call_args[0][0] == expected_error
            assert ex_info.value.code == 2


@mock.patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
@mock.patch.object(Client, "send")
class TestConsumers(CommandTestCase):
    # fmt: off
    eventhub_dsn = "Endpoint=sb://myhub.servicebus.windows.net/;SharedAccessKeyName=...;SharedAccessKey=...;EntityPath=..."  # noqa
    storage_dsn = "DefaultEndpointsProtocol=https;AccountName=...;AccountKey=...;EndpointSuffix=core.windows.net"  # noqa
    # fmt: on

    def test_deploy_consumer(self, mock_send, mock_load_config):
        project_id = gen_uuid()
        cluster_id = gen_uuid()

        argv = [
            "croud",
            "consumers",
            "deploy",
            "--product-name",
            "eventhub-consumer",
            "--tier",
            "S0",
            "--num-instances",
            "2",
            "--project-id",
            project_id,
            "--cluster-id",
            cluster_id,
            "--consumer-name",
            "my-eventhub-consumer",
            "--consumer-table",
            "raw",
            "--consumer-schema",
            "doc",
            "--eventhub-dsn",
            self.eventhub_dsn,
            "--eventhub-consumer-group",
            "$Default",
            "--lease-storage-dsn",
            self.storage_dsn,
            "--lease-storage-container",
            "lease_container",
        ]

        self.assertRest(
            mock_send,
            argv,
            RequestMethod.POST,
            "/api/v2/consumers/",
            body={
                "cluster_id": cluster_id,
                "config": {
                    "connection_string": self.eventhub_dsn,
                    "consumer_group": "$Default",
                    "lease_storage_connection_string": self.storage_dsn,
                    "consumer_lease_container": "lease_container",
                },
                "instances": 2,
                "name": "my-eventhub-consumer",
                "product_name": "eventhub-consumer",
                "product_tier": "S0",
                "project_id": project_id,
                "table_name": "raw",
                "table_schema": "doc",
            },
        )

    def test_consumers_list(self, mock_send, mock_load_config):
        argv = ["croud", "consumers", "list"]
        self.assertRest(
            mock_send, argv, RequestMethod.GET, "/api/v2/consumers/", params={}
        )

    def test_consumers_list_with_params(self, mock_send, mock_load_config):
        project_id = gen_uuid()
        cluster_id = gen_uuid()
        argv = [
            "croud",
            "consumers",
            "list",
            "--project-id",
            project_id,
            "--cluster-id",
            cluster_id,
            "--product-name",
            "eventhub-consumer",
        ]
        self.assertRest(
            mock_send,
            argv,
            RequestMethod.GET,
            "/api/v2/consumers/",
            params={
                "product_name": "eventhub-consumer",
                "project_id": project_id,
                "cluster_id": cluster_id,
            },
        )

    def test_consumers_edit(self, mock_send, mock_load_config):
        consumer_id = gen_uuid()
        cluster_id = gen_uuid()

        argv = [
            "croud",
            "consumers",
            "edit",
            "--consumer-id",
            consumer_id,
            "--eventhub-dsn",
            self.eventhub_dsn,
            "--eventhub-consumer-group",
            "$Default",
            "--lease-storage-dsn",
            self.storage_dsn,
            "--lease-storage-container",
            "lease_container",
            "--consumer-schema",
            "doc",
            "--consumer-table",
            "raw",
            "--cluster-id",
            cluster_id,
        ]

        self.assertRest(
            mock_send,
            argv,
            RequestMethod.PATCH,
            f"/api/v2/consumers/{consumer_id}/",
            body={
                "cluster_id": cluster_id,
                "config": {
                    "connection_string": self.eventhub_dsn,
                    "consumer_group": "$Default",
                    "lease_storage_connection_string": self.storage_dsn,
                    "consumer_lease_container": "lease_container",
                },
                "table_name": "raw",
                "table_schema": "doc",
            },
        )


@mock.patch("croud.config.load_config", return_value=Configuration.DEFAULT_CONFIG)
@mock.patch.object(Client, "send")
class TestGrafana(CommandTestCase):
    project_id = gen_uuid()

    def test_enable(self, mock_send, mock_config):
        argv = [
            "croud",
            "monitoring",
            "grafana",
            "enable",
            "--project-id",
            self.project_id,
        ]
        self.assertRest(
            mock_send,
            argv,
            RequestMethod.POST,
            "/api/v2/monitoring/grafana/",
            body={"project_id": self.project_id},
        )

    def test_disable(self, mock_send, mock_config):
        argv = [
            "croud",
            "monitoring",
            "grafana",
            "disable",
            "--project-id",
            self.project_id,
        ]
        self.assertRest(
            mock_send,
            argv,
            RequestMethod.DELETE,
            "/api/v2/monitoring/grafana/",
            body={"project_id": self.project_id},
        )



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import _ArgumentGroup


def project_id_arg(
    req_args: _ArgumentGroup, opt_args: _ArgumentGroup, required: bool
) -> None:
    group = req_args if required else opt_args
    group.add_argument(
        "-p", "--project-id", type=str, help="The project ID to use.", required=required
    )


def project_name_arg(req_args: _ArgumentGroup, opt_args: _ArgumentGroup) -> None:
    req_args.add_argument(
        "--name", type=str, help="The project name to use.", required=True
    )


def org_id_arg(
    req_args: _ArgumentGroup, opt_args: _ArgumentGroup, required: bool
) -> None:
    group = req_args if required else opt_args
    group.add_argument(
        "--org-id", type=str, help="The organization ID to use.", required=required
    )


def no_org_arg(req_args: _ArgumentGroup, opt_args: _ArgumentGroup) -> None:
    opt_args.add_argument(
        "--no-org",
        nargs="?",
        default=False,
        const=True,
        type=bool,
        help="Only show users that are not part of any organization.",
    )


def org_id_no_org_arg_mutual_exclusive(
    req_args: _ArgumentGroup, opt_args: _ArgumentGroup
) -> None:
    exclusive = opt_args.add_mutually_exclusive_group()
    org_id_arg(req_args, exclusive, False)
    no_org_arg(req_args, exclusive)


def org_name_arg(req_args: _ArgumentGroup, opt_args: _ArgumentGroup) -> None:
    req_args.add_argument(
        "--name", type=str, help="The organization name to use.", required=True
    )


def org_plan_type_arg(req_args: _ArgumentGroup, opt_args: _ArgumentGroup) -> None:
    req_args.add_argument(
        "--plan-type",
        choices=[1, 2, 3, 4, 5, 6],
        type=int,
        help="The support plan to use for the organization.",
        required=True,
    )


def resource_id_arg(
    req_args: _ArgumentGroup, opt_args: _ArgumentGroup, required: bool
) -> None:
    group = req_args if required else opt_args
    group.add_argument(
        "--resource", type=str, help="The resource ID to use.", required=required
    )


def role_fqn_arg(
    req_args: _ArgumentGroup, opt_args: _ArgumentGroup, required: bool
) -> None:
    group = req_args if required else opt_args
    group.add_argument(
        "--role",
        type=str,
        help=(
            "The role FQN to use. Run `croud users roles list` "
            "for a list of available roles."
        ),
        required=required,
    )


def user_id_arg(
    req_args: _ArgumentGroup, opt_args: _ArgumentGroup, required: bool
) -> None:
    group = req_args if required else opt_args
    group.add_argument(
        "--user", type=str, help="The user ID to use.", required=required
    )


def user_id_or_email_arg(req_args: _ArgumentGroup, opt_args: _ArgumentGroup) -> None:
    req_args.add_argument(
        "--user", type=str, help="The user email address or ID to use.", required=True
    )


def cluster_id_arg(
    req_args: _ArgumentGroup, opt_args: _ArgumentGroup, required: bool
) -> None:
    group = req_args if required else opt_args
    group.add_argument(
        "--cluster-id",
        type=str,
        help="The CrateDB cluster ID to use.",
        required=required,
    )


def cluster_name_arg(
    req_args: _ArgumentGroup, opt_args: _ArgumentGroup, required: bool
) -> None:
    group = req_args if required else opt_args
    group.add_argument(
        "--cluster-name",
        type=str,
        help="The CrateDB cluster name to use.",
        required=required,
    )


def product_tier_arg(req_args: _ArgumentGroup, opt_args: _ArgumentGroup) -> None:
    req_args.add_argument(
        "--tier", type=str, help="The product tier to use.", required=True
    )


def product_unit_arg(
    req_args: _ArgumentGroup, opt_args: _ArgumentGroup, required: bool
) -> None:
    group = req_args if required else opt_args
    group.add_argument(
        "--unit", type=int, help="The product scale unit to use.", required=required
    )


def product_name_arg(
    req_args: _ArgumentGroup, opt_args: _ArgumentGroup, required: bool
) -> None:
    group = req_args if required else opt_args
    group.add_argument(
        "--product-name", type=str, help="The product name to use.", required=required
    )


def crate_version_arg(req_args: _ArgumentGroup, opt_args: _ArgumentGroup) -> None:
    req_args.add_argument(
        "--version", type=str, help="The CrateDB version to use.", required=True
    )


def crate_username_arg(req_args: _ArgumentGroup, opt_args: _ArgumentGroup) -> None:
    req_args.add_argument(
        "--username", type=str, help="The CrateDB username to use.", required=True
    )


def crate_password_arg(req_args: _ArgumentGroup, opt_args: _ArgumentGroup) -> None:
    req_args.add_argument(
        "--password", type=str, help="The CrateDB password to use.", required=True
    )


def consumer_id_arg(req_args: _ArgumentGroup, opt_args: _ArgumentGroup) -> None:
    req_args.add_argument(
        "--consumer-id", type=str, help="The consumer set ID to use.", required=True
    )


def consumer_name_arg(req_args: _ArgumentGroup, opt_args: _ArgumentGroup) -> None:
    req_args.add_argument(
        "--consumer-name", type=str, help="The consumer name to use.", required=True
    )


def consumer_schema_arg(
    req_args: _ArgumentGroup, opt_args: _ArgumentGroup, required: bool = True
) -> None:
    group = req_args if required else opt_args
    group.add_argument(
        "--consumer-schema",
        type=str,
        help="The CrateDB database schema used by the Azure EventHub consumer.",
        required=required,
    )


def consumer_table_arg(
    req_args: _ArgumentGroup, opt_args: _ArgumentGroup, required: bool = True
) -> None:
    group = req_args if required else opt_args
    group.add_argument(
        "--consumer-table",
        type=str,
        help="The CrateDB database table used by the Azure EventHub consumer.",
        required=required,
    )


def eventhub_dsn_arg(
    req_args: _ArgumentGroup, opt_args: _ArgumentGroup, required: bool = True
) -> None:
    group = req_args if required else opt_args
    group.add_argument(
        "--eventhub-dsn",
        type=str,
        help="The connection string to the Azure EventHub from which to consume.",
        required=required,
    )


def eventhub_consumer_group_arg(
    req_args: _ArgumentGroup, opt_args: _ArgumentGroup, required: bool = True
) -> None:
    group = req_args if required else opt_args
    group.add_argument(
        "--eventhub-consumer-group",
        type=str,
        help="The consumer group of the Azure EventHub from which to consume.",
        required=required,
    )


def lease_storage_dsn_arg(
    req_args: _ArgumentGroup, opt_args: _ArgumentGroup, required: bool = True
) -> None:
    group = req_args if required else opt_args
    group.add_argument(
        "--lease-storage-dsn",
        type=str,
        help=(
            "The connection string to an Azure storage account to use as lease storage."
        ),
        required=required,
    )


def lease_storage_container_arg(
    req_args: _ArgumentGroup, opt_args: _ArgumentGroup, required: bool = True
) -> None:
    group = req_args if required else opt_args
    group.add_argument(
        "--lease-storage-container",
        type=str,
        help=(
            "The container name in the lease storage for the Azure EventHub "
            "consumer to use."
        ),
        required=required,
    )


def num_instances_arg(req_args: _ArgumentGroup, opt_args: _ArgumentGroup) -> None:
    opt_args.add_argument(
        "--num-instances",
        type=int,
        help="The number of instances to deploy.",
        default=1,
        required=False,
    )



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import Namespace

from croud.rest import Client
from croud.session import RequestMethod


def clusters_list(args: Namespace) -> None:
    """
    Lists all projects for the current user in the specified region
    """

    params = {}
    if args.project_id:
        params["project_id"] = args.project_id

    client = Client(env=args.env, region=args.region, output_fmt=args.output_fmt)
    client.send(RequestMethod.GET, "/api/v2/clusters/", params=params)
    client.print(
        keys=[
            "id",
            "name",
            "num_nodes",
            "crate_version",
            "project_id",
            "username",
            "fqdn",
        ]
    )


def clusters_deploy(args: Namespace) -> None:
    """
    Deploys a new cluster CrateDB cluster.
    """
    body = {
        "crate_version": args.version,
        "name": args.cluster_name,
        "password": args.password,
        "product_name": args.product_name,
        "product_tier": args.tier,
        "project_id": args.project_id,
        "username": args.username,
    }
    if args.unit:
        body["product_unit"] = args.unit
    client = Client(env=args.env, region=args.region, output_fmt=args.output_fmt)
    client.send(RequestMethod.POST, "/api/v2/clusters/", body=body)
    client.print(keys=["id", "name", "fqdn", "url"])



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from http.cookies import Morsel
from unittest import mock

import aiohttp
import pytest
from aiohttp.test_utils import loop_context
from util.fake_server import FakeCrateDBCloud, FakeResolver

from croud.config import Configuration
from croud.session import HttpSession, RequestMethod, cloud_url

me_query = """
{
  me {
    email
    username
    name
  }
}
"""

me_response = {
    "me": {
        "email": "sheldon@crate.io",
        "username": "Google_1234",
        "name": "Sheldon Cooper",
    }
}


@mock.patch.object(Configuration, "get_env", return_value="dev")
class TestHttpSession:
    @pytest.mark.parametrize("variables", [None, {"a": "foo", "b": 42}])
    @mock.patch.object(Configuration, "get_token", return_value="eyJraWQiOiIx")
    def test_query_success(self, mock_token, mock_env, variables):
        with loop_context() as loop:

            # Takes a query that will be used to determine which response is
            # generated by the FakeCrateDBCloud instance
            async def test_query(headers_query: dict):
                env = Configuration.get_env()
                token = Configuration.get_token()
                async with HttpSession(
                    env,
                    token,
                    url="https://cratedb.local",
                    conn=connector,
                    headers=headers_query,  # Determines the generated response
                ) as session:
                    resp = await session.fetch(
                        RequestMethod.POST,
                        "/graphql",
                        {"query": me_query, "variables": variables},
                    )
                    assert resp.status == 200

                    data = await resp.json()
                    assert data["data"] == me_response

            fake_cloud = FakeCrateDBCloud(loop=loop)
            info = loop.run_until_complete(fake_cloud.start())
            resolver = FakeResolver(info, loop=loop)
            connector = aiohttp.TCPConnector(loop=loop, resolver=resolver, ssl=True)

            query = {"query": "me"}
            loop.run_until_complete(test_query(query))
            loop.run_until_complete(fake_cloud.stop())

    @mock.patch.object(Configuration, "get_token", return_value="")
    @mock.patch("croud.session.print_error")
    def test_request_unauthorized(self, mock_print_error, mock_token, mock_env):
        with loop_context() as loop:

            # Takes a query that will be used to determine which response is
            # generated by the FakeCrateDBCloud instance
            async def test_request(headers_query: dict):
                env = Configuration.get_env()
                token = Configuration.get_token()
                with pytest.raises(SystemExit) as cm:
                    async with HttpSession(
                        env,
                        token,
                        url="https://cratedb.local",
                        conn=connector,
                        headers=headers_query,  # Determines the generated response
                    ) as session:
                        await session.fetch(
                            RequestMethod.POST,
                            "/graphql",
                            {"query": me_query, "variables": None},
                        )
                mock_print_error.assert_called_once_with(
                    "Unauthorized. Use `croud login` to login to CrateDB Cloud."
                )
                assert cm.value.code == 1

            fake_cloud = FakeCrateDBCloud(loop=loop)
            info = loop.run_until_complete(fake_cloud.start())
            resolver = FakeResolver(info, loop=loop)
            connector = aiohttp.TCPConnector(loop=loop, resolver=resolver, ssl=True)

            query = {"query": "me"}
            loop.run_until_complete(test_request(query))
            loop.run_until_complete(fake_cloud.stop())

    def test_correct_url(self, mock_env):
        region = "bregenz.a1"

        url = cloud_url("prod", region)
        assert url == f"https://{region}.cratedb.cloud"

        url = cloud_url("dev", region)
        assert url == f"https://{region}.cratedb-dev.cloud"

        url = cloud_url("local", region)
        assert url == "http://localhost:8000"

        region = "westeurope.azure"
        url = cloud_url("prod", region)
        assert url == f"https://{region}.cratedb.cloud"


@pytest.mark.asyncio
async def test_on_new_token():
    update_config = mock.Mock()
    async with HttpSession(
        "dev", "old_token", "eastus.azure", on_new_token=update_config
    ) as session:
        session_cookie = Morsel()
        session_cookie.set("session", "new_token", None)
        session.client.cookie_jar.update_cookies({"session": session_cookie})

    update_config.assert_called_once_with("new_token")



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import os
import platform
import subprocess
import webbrowser
from typing import Dict, Tuple


# This function was copied from the <https://github.com/Azure/azure-cli>
# project. See `LICENSE` for more information.
def can_launch_browser() -> bool:
    platform_name, release = get_platform_info()
    if is_wsl(platform_name, release) or platform_name != "linux":
        return True
    gui_env_vars = ["DESKTOP_SESSION", "XDG_CURRENT_DESKTOP", "DISPLAY"]
    result = True
    if platform_name == "linux":
        if any(os.getenv(v) for v in gui_env_vars):
            try:
                default_browser = webbrowser.get()
                if (
                    getattr(default_browser, "name", None) == "www-browser"
                ):  # text browser won't work
                    result = False
            except webbrowser.Error:
                result = False
        else:
            result = False

    return result


# This function was copied from the <https://github.com/Azure/azure-cli>
# project. See `LICENSE` for more information.
def is_wsl(platform_name: str, release: str) -> bool:
    platform_name, release = get_platform_info()
    return platform_name == "linux" and release.split("-")[-1] == "microsoft"


# This function was copied from the <https://github.com/Azure/azure-cli>
# project. See `LICENSE` for more information.
def get_platform_info() -> Tuple[str, str]:
    uname = platform.uname()
    platform_name = getattr(uname, "system", None) or uname[0]
    release = getattr(uname, "release", None) or uname[2]
    return platform_name.lower(), release.lower()


# This function was copied from the <https://github.com/Azure/azure-cli>
# project. See `LICENSE` for more information.
def open_page_in_browser(url: str) -> int:
    platform_name, release = get_platform_info()
    if is_wsl(platform_name, release):  # windows 10 linux subsystem
        try:
            return subprocess.call(
                ["cmd.exe", "/c", "start {}".format(url.replace("&", "^&"))]
            )
        except FileNotFoundError:  # WSL might be too old
            pass
    return webbrowser.open_new_tab(url)


def clean_dict(data: Dict) -> Dict:
    """Clean a dictionary recursively from keys with None values.
    Returns a new instance of a dict.
    """
    cleaned = {}
    for k, v in data.items():
        if v is None:
            continue
        elif isinstance(v, dict):
            cleaned[k] = clean_dict(v)
        else:
            cleaned[k] = v
    return cleaned



import argparse
import difflib
import re
from doctest import _ellipsis_match  # type: ignore
from functools import partial

from croud.__main__ import get_parser

normalize = partial(re.sub, r"\s+", "")


def ndiff(actual: str, expected: str) -> str:
    engine = difflib.Differ(charjunk=difflib.IS_CHARACTER_JUNK)
    diff = list(engine.compare(expected.splitlines(True), actual.splitlines(True)))
    return "\n".join(line.rstrip() for line in diff)


def assert_ellipsis_match(actual: str, expected: str) -> None:
    diff = ndiff(expected.strip(), actual.strip())
    equality = _ellipsis_match(normalize(expected.strip()), normalize(actual.strip()))
    assert equality, diff


class CommandTestCase:
    parser = get_parser()

    def parse(self, argv) -> argparse.Namespace:
        return self.parser.parse_args(argv[1:])

    def execute(self, argv) -> None:
        params = self.parse(argv)
        if "resolver" in params:
            fn = params.resolver
            del params.resolver
            fn(params)
        else:
            self.parser.print_help()

    def assertGql(self, mock_run, argv, expected_body, expected_vars=None):
        self.execute(argv)
        mock_run.assert_called_once_with(expected_body, expected_vars)

    def assertRest(self, mock_send, argv, method, endpoint, *, body=None, params=None):
        self.execute(argv)

        args = [method, endpoint]
        kwargs = {}
        if body is not None:
            kwargs["body"] = body
        if params is not None:
            kwargs["params"] = params
        mock_send.assert_called_once_with(*args, **kwargs)



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import asyncio
from argparse import Namespace
from typing import Dict, Optional

from croud.config import Configuration
from croud.printer import print_error, print_format, print_info, print_success
from croud.session import HttpSession, RequestMethod
from croud.typing import JsonDict

DEFAULT_ENDPOINT = "/graphql"


class Query:
    def __init__(self, query: str, args: Namespace, endpoint=DEFAULT_ENDPOINT) -> None:
        self._query = query
        self._token = Configuration.get_token()

        self._env = args.env or Configuration.get_env()
        self._output_fmt = (
            hasattr(args, "output_fmt")  # certain commands do not have an output format
            and args.output_fmt
            or Configuration.get_setting("output_fmt")
        )
        self._region = (
            hasattr(args, "region")  # certain commands do not have region
            and args.region
            or Configuration.get_setting("region")
        )
        self._endpoint = endpoint

        self._error: Optional[str] = None
        self._response: Optional[JsonDict] = None

    async def _fetch_data(self, body: str, variables: Optional[Dict]) -> JsonDict:
        async with HttpSession(
            self._env, self._token, self._region, on_new_token=Configuration.set_token
        ) as session:
            resp = await session.fetch(
                RequestMethod.POST,
                self._endpoint,
                {"query": body, "variables": variables},
            )

            if resp.status != 200:
                print_info(f"Query failed to run by returning code of {resp.status}.")
                print_info(body)
                if variables:
                    print_info(str(variables))

            return await resp.json()

    def run(self, body: str, variables: Optional[Dict]) -> JsonDict:
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(self._fetch_data(body, variables))

    def execute(self, variables: Dict = None):
        response = self.run(self._query, variables)

        if "errors" in response:
            self._response = None
            self._error = response["errors"][0]["message"]
        elif "data" in response:
            self._response = response["data"]
            self._error = None
        else:
            self._response = None
            self._error = None


def print_query(query: Query, key: str = None, success_message: str = None) -> None:
    if query._error:
        print_error(query._error)
        return

    if not query._response:
        print_info("Result contained no data to print.")
        return

    if not isinstance(query._response, dict):
        print_error("Result has no proper format to print.")
        return

    if key:
        data = query._response[key]
    else:
        data = query._response

    if "success" in data:
        success = data["success"]
        if success:
            if success_message:
                print_success(success_message)
            else:
                print_success("Success.")
        else:
            # Edge case that might occur during network partitions/timeouts etc.
            print_error(
                "Command not successful, however no server-side errors occurred. "
                "Please try again."
            )
        return

    # some queries have two levels to access data
    # e.g. {allProjects: {data: [{projects}]}} vs. {me: {data})
    if "data" in data:
        data = data["data"]

    if len(data) == 0:
        print_info("Result contained no data to print.")
    else:
        print_format(data, query._output_fmt)



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import os
import tempfile
import unittest
import uuid
from unittest import mock

from yaml.constructor import ConstructorError

from croud.config import Configuration, load_config, write_config


class TestConfiguration(unittest.TestCase):
    @mock.patch.object(Configuration, "validate", return_value={})
    @mock.patch("yaml.safe_load")
    def test_load_config(self, mock_yaml_load, mock_validate):
        m = mock.mock_open()
        with mock.patch("croud.config.open", m, create=True):
            load_config()

        mock_yaml_load.assert_called_once_with(m())

    @mock.patch("yaml.dump")
    def test_write_config(self, mock_yaml_dump):
        config = Configuration.DEFAULT_CONFIG
        m = mock.mock_open()
        with mock.patch("croud.config.open", m, create=True):
            write_config(config)

        mock_yaml_dump.assert_called_once_with(
            config, m(), default_flow_style=False, allow_unicode=True
        )

    @mock.patch("croud.config.os.remove")
    @mock.patch("croud.config.os.chmod")
    @mock.patch("croud.config.write_config")
    def test_incompatible_schema_rewrites_config(
        self, mock_write_config, mock_chmod, mock_remove
    ):
        config = {"incompatible": "", "schema": ""}

        m = mock.mock_open()
        with mock.patch("croud.config.open", m, create=True):
            Configuration.validate(config)
            self.assertTrue(mock_remove.called)

            mock_write_config.assert_called_once_with(Configuration.DEFAULT_CONFIG)

    def test_yaml_safe_load(self):
        new_file = os.path.join(tempfile.gettempdir(), uuid.uuid4().hex)
        with tempfile.NamedTemporaryFile(mode="w+") as tmp:
            tmp.write(
                "insecure: !!python/object/apply:"
                f"subprocess.run [['touch', '{new_file}']]"
            )
            tmp.flush()
            self.assertFalse(os.path.exists(new_file))
            with mock.patch.object(Configuration, "FILEPATH", tmp.name):
                with self.assertRaisesRegex(
                    ConstructorError,
                    r"tag:yaml.org,\d+:python/object/apply:subprocess.run",
                ):
                    load_config()

            self.assertFalse(os.path.exists(new_file))



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import Namespace

from croud.rest import Client
from croud.session import RequestMethod


def consumers_deploy(args: Namespace) -> None:
    body = {
        "cluster_id": args.cluster_id,
        "config": {
            "connection_string": args.eventhub_dsn,
            "consumer_group": args.eventhub_consumer_group,
            "consumer_lease_container": args.lease_storage_container,
            "lease_storage_connection_string": args.lease_storage_dsn,
        },
        "instances": args.num_instances,
        "name": args.consumer_name,
        "product_name": args.product_name,
        "product_tier": args.tier,
        "project_id": args.project_id,
        "table_name": args.consumer_table,
        "table_schema": args.consumer_schema,
    }
    client = Client(env=args.env, region=args.region, output_fmt=args.output_fmt)
    client.send(RequestMethod.POST, "/api/v2/consumers/", body=body)
    client.print(
        keys=[
            "cluster_id",
            "id",
            "instances",
            "name",
            "product_name",
            "product_tier",
            "project_id",
            "table_name",
            "table_schema",
        ]
    )


def consumers_list(args: Namespace) -> None:
    params = {}
    if args.cluster_id:
        params["cluster_id"] = args.cluster_id
    if args.product_name:
        params["product_name"] = args.product_name
    if args.project_id:
        params["project_id"] = args.project_id

    client = Client(env=args.env, region=args.region, output_fmt=args.output_fmt)
    client.send(RequestMethod.GET, "/api/v2/consumers/", params=params)
    client.print(
        keys=[
            "cluster_id",
            "id",
            "instances",
            "name",
            "product_name",
            "product_tier",
            "project_id",
            "table_name",
            "table_schema",
        ]
    )


def consumers_edit(args: Namespace) -> None:
    body = {
        "cluster_id": args.cluster_id,
        "table_name": args.consumer_table,
        "table_schema": args.consumer_schema,
    }
    config = {}
    if args.eventhub_dsn:
        config["connection_string"] = args.eventhub_dsn
    if args.eventhub_consumer_group:
        config["consumer_group"] = args.eventhub_consumer_group
    if args.lease_storage_container:
        config["consumer_lease_container"] = args.lease_storage_container
    if args.lease_storage_dsn:
        config["lease_storage_connection_string"] = args.lease_storage_dsn

    if config:
        body["config"] = config
    client = Client(env=args.env, region=args.region, output_fmt=args.output_fmt)
    client.send(
        RequestMethod.PATCH, f"/api/v2/consumers/{args.consumer_id}/", body=body
    )
    client.print(keys=["id"])



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import asyncio
from typing import List, Optional, Union

from aiohttp import ContentTypeError  # type: ignore
from aiohttp.client_reqrep import ClientResponse

from croud.config import Configuration
from croud.printer import print_error, print_format, print_success
from croud.session import HttpSession, RequestMethod


class Client:
    _env: str
    _token: str
    _region: str
    _output_fmt: str
    _data: Optional[Union[List[dict], dict]] = None
    _error: Optional[dict] = None

    def __init__(
        self, env: str = None, region: str = None, output_fmt: str = None, loop=None
    ):
        self._env = env or Configuration.get_env()
        self._region = region or Configuration.get_setting("region")
        self._output_fmt = output_fmt or Configuration.get_setting("output_fmt")
        self.loop = loop or asyncio.get_event_loop()

    def send(
        self,
        method: RequestMethod,
        endpoint: str,
        *,
        body: dict = None,
        params: dict = None
    ):
        self.loop.run_until_complete(self.fetch(method, endpoint, body, params))

    async def fetch(
        self,
        method: RequestMethod,
        endpoint: str,
        body: dict = None,
        params: dict = None,
    ):
        resp = await self._fetch(method, endpoint, body, params)
        data = await self._decode_response(resp)

        if resp.status >= 400:
            self._error = data
            return

        self._data = data["data"] if "data" in data else data

    async def _fetch(
        self,
        method: RequestMethod,
        endpoint: str,
        body: dict = None,
        params: dict = None,
    ) -> ClientResponse:
        async with HttpSession(
            self._env,
            Configuration.get_token(),
            self._region,
            on_new_token=Configuration.set_token,
        ) as session:
            return await session.fetch(method, endpoint, body, params)

    async def _decode_response(self, resp: ClientResponse):
        try:
            return await resp.json()
        # API always returns JSON, unless there's an unhandled server error
        except ContentTypeError:
            return {"message": "Invalid response type.", "success": False}

    def print(self, success_message: str = None, keys: List[str] = None):
        if self._error:
            if "message" in self._error:
                print_error(self._error["message"])
            else:
                print_format(self._error, "json")
            return

        if self._data is None or len(self._data) == 0:
            message = success_message or "Success."
            print_success(message)
            return

        print_format(self._data, self._output_fmt, keys)



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

import os
from argparse import Namespace

import yaml
from appdirs import user_config_dir
from schema import Schema, SchemaError

from croud.printer import print_error, print_format, print_info


class IncompatibleConfigException(Exception):
    pass


class Configuration:
    USER_CONFIG_DIR: str = user_config_dir("Crate")
    FILENAME: str = "croud.yaml"
    FILEPATH: str = f"{USER_CONFIG_DIR}/{FILENAME}"
    DEFAULT_CONFIG: dict = {
        "auth": {
            "current_context": "prod",
            "contexts": {
                "prod": {"token": ""},
                "dev": {"token": ""},
                "local": {"token": ""},
            },
        },
        "region": "bregenz.a1",
        "output_fmt": "table",
    }
    CONFIG_NAMES: dict = {
        "env": "Environment",
        "output_fmt": "Output format",
        "region": "Region",
    }

    current_context: str = ""

    @staticmethod
    def validate(config: dict) -> dict:
        schema = Schema(
            {
                "auth": {
                    "current_context": str,
                    "contexts": {
                        "prod": {"token": str},
                        "dev": {"token": str},
                        "local": {"token": str},
                    },
                },
                "region": str,
                "output_fmt": str,
            }
        )
        try:
            return schema.validate(config)
        except SchemaError:
            os.remove(Configuration.FILEPATH)
            create_default_config()

            return schema.validate(Configuration.DEFAULT_CONFIG)

    @staticmethod
    def create() -> None:
        if not os.path.exists(Configuration.USER_CONFIG_DIR):
            os.makedirs(Configuration.USER_CONFIG_DIR)

        if not os.path.isfile(Configuration.FILEPATH):
            create_default_config()
        else:
            load_config()

    @staticmethod
    def get_setting(setting: str) -> str:
        setting = setting.replace("-", "_")
        if setting == "env":
            return Configuration.get_env()
        return load_config()[setting]

    @staticmethod
    def get_env() -> str:
        if Configuration.current_context:
            return Configuration.current_context

        config = load_config()
        return config["auth"]["current_context"]

    @staticmethod
    def set_context(env: str) -> None:
        config = load_config()
        config["auth"]["current_context"] = env
        write_config(config)

    @staticmethod
    def get_token() -> str:
        return get_auth_context().get("token", "")

    @staticmethod
    def set_token(token: str) -> None:
        config = load_config()

        context = Configuration.current_context
        if not context:
            context = config["auth"]["current_context"]

        config["auth"]["contexts"][context]["token"] = token

        write_config(config)

    @staticmethod
    def override_context(env: str) -> None:
        Configuration.current_context = env


def create_default_config() -> None:
    write_config(Configuration.DEFAULT_CONFIG)
    os.chmod(Configuration.FILEPATH, 0o600)


def load_config() -> dict:
    with open(Configuration.FILEPATH, "r") as f:
        config = yaml.safe_load(f)

    try:
        return Configuration.validate(config)
    except IncompatibleConfigException as e:
        print_error(str(e))
        exit(1)


def set_property(property: str, value: str):
    config = load_config()
    config[property] = value
    write_config(config)


def write_config(config: dict) -> None:
    with open(Configuration.FILEPATH, "w", encoding="utf8") as f:
        yaml.dump(config, f, default_flow_style=False, allow_unicode=True)


def get_auth_context() -> dict:
    config = load_config()

    context = Configuration.current_context
    if not context:
        context = config["auth"]["current_context"]

    return config["auth"]["contexts"][context]


def config_get(args: Namespace):
    """
    Gets a default configuration setting
    """

    fmt = Configuration.get_setting("output_fmt")
    value = Configuration.get_setting(args.get)
    print_format([{args.get: value}], args.output_fmt or fmt)


def config_set(args: Namespace):
    """
    Sets a default configuration setting
    """

    for key in vars(args):
        setting = getattr(args, key)

        if setting is not None:
            if key == "env":
                Configuration.set_context(setting)
                print_info(f"Environment switched to {setting}")
            else:
                set_property(key, setting)
                print_info(f"{Configuration.CONFIG_NAMES[key]} switched to {setting}")



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.


import argparse
import sys

from croud import __version__

POSITIONALS_TITLE = "Available Commands"
REQUIRED_TITLE = "Required Arguments"
OPTIONALS_TITLE = "Optional Arguments"


class CroudCliArgumentParser(argparse.ArgumentParser):
    def __init__(self, **kwargs):
        super().__init__(
            formatter_class=CroudCliHelpFormatter, add_help=False, **kwargs
        )
        self._positionals.title = POSITIONALS_TITLE
        self._optionals.title = OPTIONALS_TITLE

        self._group_required = self.add_argument_group(REQUIRED_TITLE)
        self._group_optional = self.add_argument_group(OPTIONALS_TITLE)
        self._group_optional.add_argument(
            "-h",
            "--help",
            action="help",
            default=argparse.SUPPRESS,
            help="Show this help message and exit.",
        )

    def error(self, message):
        sys.stderr.write(f"{message.capitalize()}\n\n")
        self.print_help()
        sys.exit(2)


class CroudCliHelpFormatter(argparse.HelpFormatter):
    def _format_action(self, action):
        result = super()._format_action(action)
        if isinstance(action, argparse._SubParsersAction):
            # Remove empty line
            return "%*s%s" % (self._current_indent, "", result.lstrip())
        return result

    def _format_action_invocation(self, action):
        if isinstance(action, argparse._SubParsersAction):
            # Remove metavar {command}
            return ""
        return super()._format_action_invocation(action)

    def _iter_indented_subactions(self, action):
        if isinstance(action, argparse._SubParsersAction):
            try:
                get_subactions = action._get_subactions
            except AttributeError:
                pass
            else:
                # Remove indentation
                yield from get_subactions()
        else:
            yield from super()._iter_indented_subactions(action)

    def add_usage(self, usage, actions, groups, prefix=None):
        if prefix is None:
            prefix = "Usage: "

        return super(CroudCliHelpFormatter, self).add_usage(
            usage, actions, groups, prefix
        )


def add_default_args(parser):
    parser._group_optional.add_argument(
        "--region",
        "-r",
        required=False,
        choices={"eastus.azure", "eastus2.azure", "westeurope.azure", "bregenz.a1"},
        help="Temporarily use the specified region that command will be run in.",
    )
    parser._group_optional.add_argument(
        "--env",
        "-e",
        required=False,
        choices={"dev", "prod", "local"},
        help="Switches auth context.",
    )
    parser._group_optional.add_argument(
        "--output-fmt",
        "-o",
        required=False,
        choices={"table", "json"},
        help="Change the formatting of the output",
    )


def help_print_factory(parser: argparse.ArgumentParser):
    def print_help(*args, **kwargs):
        parser.print_help()

    return print_help


def add_subparser(parser, tree, name="__root__"):
    if "extra_args" in tree:
        for arg_provider in tree["extra_args"]:
            arg_provider(
                parser._group_required, parser._group_optional
            )  # keep existing behaviour
    if "noop_arg" in tree:
        parser.add_argument(name.split(" ")[-1], choices=tree["noop_arg"]["choices"])
    if "commands" in tree:
        parser.set_defaults(resolver=help_print_factory(parser))
        subparsers = parser.add_subparsers()
        for _cmd, _tree in tree["commands"].items():
            sub = subparsers.add_parser(_cmd, help=_tree.get("help"))
            add_subparser(sub, _tree, sub.prog)
    else:
        add_default_args(parser)
        parser.set_defaults(resolver=tree["resolver"])


def create_parser(tree):
    parser = CroudCliArgumentParser(prog="croud", description=tree["help"])
    parser._group_optional.add_argument(
        "-v",
        "--version",
        action="version",
        version="%(prog)s " + __version__,
        help="Show program's version number and exit.",
    )
    add_subparser(parser, tree, parser.prog)
    return parser



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import Namespace

from croud.gql import Query, print_query


def project_user_add(args: Namespace) -> None:
    """
    Adds a user to a project.
    """

    _query = f"""
    mutation {{
        addUserToProject(input: {{
            projectId: "{args.project_id}",
            user: "{args.user}"
        }}) {{
            success
        }}
    }}
    """

    query = Query(_query, args)
    query.execute()
    print_query(query, "addUserToProject", "Successfully added user to project.")


def project_user_remove(args: Namespace) -> None:
    """
    Removes a user to a project.
    """

    _query = f"""
    mutation {{
        removeUserFromProject(input: {{
            projectId: "{args.project_id}",
            user: "{args.user}"
        }}) {{
            success
        }}
    }}
    """

    query = Query(_query, args)
    query.execute()
    print_query(
        query, "removeUserFromProject", "Successfully removed user from project."
    )



# Licensed to CRATE Technology GmbH ("Crate") under one or more contributor
# license agreements.  See the NOTICE file distributed with this work for
# additional information regarding copyright ownership.  Crate licenses
# this file to you under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.  You may
# obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# License for the specific language governing permissions and limitations
# under the License.
#
# However, if you have executed another commercial license agreement
# with Crate these terms will supersede the license and you may use the
# software solely pursuant to the terms of the relevant commercial agreement.

from argparse import Namespace
from typing import Dict
from unittest import mock

import pytest

from croud.config import Configuration
from croud.login import _login_url, _set_login_env, login
from croud.logout import _logout_url, logout
from croud.server import Server


class MockConfig:
    """
    A mocked configuration which emulates reading and writing file from/to disk.
    """

    def __init__(self, config: Dict) -> None:
        self.conf = config

    def read_config(self) -> Dict:
        return self.conf

    def write_config(self, config) -> None:
        self.conf = config


class TestLogin:
    @mock.patch.object(Server, "stop")
    @mock.patch.object(Server, "start")
    @mock.patch("croud.login.asyncio.get_event_loop")
    @mock.patch("croud.login.can_launch_browser", return_value=True)
    @mock.patch("croud.login.open_page_in_browser")
    @mock.patch("croud.login.print_info")
    def test_login_success(
        self,
        mock_print_info,
        mock_open_page_in_browser,
        mock_can_launch_browser,
        mock_loop,
        mock_start,
        mock_stop,
    ):
        cfg = MockConfig(Configuration.DEFAULT_CONFIG)

        with mock.patch("croud.config.load_config", side_effect=cfg.read_config):
            with mock.patch("croud.config.write_config", side_effect=cfg.write_config):
                login(Namespace(env="dev"))

        calls = [
            mock.call("A browser tab has been launched for you to login."),
            mock.call("Login successful."),
        ]
        mock_print_info.assert_has_calls(calls)
        assert cfg.read_config()["auth"]["current_context"] == "dev"

    @mock.patch.object(Server, "stop")
    @mock.patch.object(Server, "start")
    @mock.patch("croud.login.asyncio.get_event_loop")
    @mock.patch("croud.login.can_launch_browser", return_value=True)
    @mock.patch("croud.login.open_page_in_browser")
    @mock.patch("croud.login.print_info")
    def test_local_login(
        self,
        mock_print_info,
        mock_open_page_in_browser,
        mock_can_launch_browser,
        mock_loop,
        mock_start,
        mock_stop,
    ):
        """
        Test for a bug that caused that upon login to local env the local token
        was overwritten by the dev token.
        """
        conf = {
            "auth": {
                "current_context": "prod",
                "contexts": {
                    "prod": {"token": "prod-token"},
                    "dev": {"token": "dev-token"},
                    "local": {"token": ""},
                },
            },
            "region": "bregenz.a1",
            "output_fmt": "table",
        }
        cfg = MockConfig(conf)

        with mock.patch("croud.config.load_config", side_effect=cfg.read_config):
            with mock.patch("croud.config.write_config", side_effect=cfg.write_config):
                login(Namespace(env="local"))

        new_cfg = cfg.read_config()
        assert new_cfg["auth"]["current_context"] == "local"
        assert (
            new_cfg["auth"]["contexts"]["local"]["token"]
            != new_cfg["auth"]["contexts"]["dev"]["token"]
        )
        assert (
            new_cfg["auth"]["contexts"]["local"]["token"]
            != new_cfg["auth"]["contexts"]["prod"]["token"]
        )

    @mock.patch("croud.config.Configuration.get_env", return_value="dev")
    @mock.patch("croud.login.can_launch_browser", return_value=False)
    @mock.patch("croud.login.print_error")
    def test_login_no_valid_browser(
        self, mock_print_error, mock_can_launch_browser, mock_get_env
    ):
        with pytest.raises(SystemExit) as e_info:
            login(Namespace(env=None))

        mock_print_error.assert_called_once_with(
            "Login only works with a valid browser installed."
        )
        assert e_info.value.code == 1

    @mock.patch("croud.config.Configuration.override_context")
    @mock.patch("croud.config.Configuration.get_env", return_value="dev")
    def test_login_env_from_current_context(self, mock_get_env, mock_override_context):
        env = _set_login_env(None)
        assert env == "dev"

    def test_login_env_override_context_from_argument(self):
        env = _set_login_env("prod")
        assert env == "prod"

    def test_login_urls_from_valid_envs(self):
        url = _login_url("dev")
        assert "https://bregenz.a1.cratedb-dev.cloud/oauth2/login?cli=true" == url

        url = _login_url("prod")
        assert "https://bregenz.a1.cratedb.cloud/oauth2/login?cli=true" == url

        url = _login_url("PROD")
        assert "https://bregenz.a1.cratedb.cloud/oauth2/login?cli=true" == url

        url = _login_url("local")
        assert "http://localhost:8000/oauth2/login?cli=true" == url

    def test_env_fallback_url(self):
        url = _login_url("invalid")
        assert "https://bregenz.a1.cratedb.cloud/oauth2/login?cli=true" == url


class TestLogout:
    @mock.patch("croud.logout.asyncio.get_event_loop")
    @mock.patch("croud.logout.print_info")
    def test_logout(self, mock_print_info, mock_loop):
        conf = {
            "auth": {
                "current_context": "prod",
                "contexts": {"prod": {"token": "my-token"}},
            },
            "region": "bregenz.a1",
            "output_fmt": "table",
        }
        cfg = MockConfig(conf)

        with mock.patch("croud.config.load_config", side_effect=cfg.read_config):
            with mock.patch("croud.config.write_config", side_effect=cfg.write_config):
                logout(Namespace(env="prod"))

        assert cfg.read_config()["auth"]["contexts"]["prod"]["token"] == ""
        mock_print_info.assert_called_once_with("You have been logged out.")

    def test_logout_urls_from_valid_envs(self):
        url = _logout_url("dev")
        assert "https://bregenz.a1.cratedb-dev.cloud/oauth2/logout" == url

        url = _logout_url("prod")
        assert "https://bregenz.a1.cratedb.cloud/oauth2/logout" == url

        url = _logout_url("PROD")
        assert "https://bregenz.a1.cratedb.cloud/oauth2/logout" == url

        url = _logout_url("local")
        assert "http://localhost:8000/oauth2/logout" == url

    def test_env_fallback_url(self):
        url = _logout_url("invalid")
        assert "https://bregenz.a1.cratedb.cloud/oauth2/logout" == url
